# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from argparse import Namespace, ArgumentParser, FileType
from datetime import date
from pathlib import Path
import re
import sys
import subprocess
from typing import Callable, Dict, List, Union

import tomlkit
from tomlkit.toml_document import TOMLDocument

from pontos.terminal import _set_terminal, error, info, out, warning
from pontos.terminal.terminal import Terminal
from pontos.release.helper import (
    get_project_name,
)


class ChangelogBuilder:
    """Creates Changelog files from conventional commits
    using the git log, from the latest tag"""

    def __init__(
        self,
        shell_cmd_runner: Callable,
        args: Namespace,
    ):
        self.shell_cmd_runner: Callable = shell_cmd_runner
        self.config: TOMLDocument = tomlkit.parse(
            args.config.read_text(encoding='utf-8')
        )
        self.project: str = (
            args.project
            if args.project is not None
            else get_project_name(self.shell_cmd_runner)
        )
        self.space: str = args.space
        changelog_dir: Path = Path.cwd() / self.config.get('changelog_dir')
        changelog_dir.mkdir(parents=True, exist_ok=True)
        self.output_file: Path = changelog_dir / args.output
        self.current_version: str = args.current_version
        self.next_version: str = args.next_version

    def create_changelog_file(self) -> Union[Path, None]:
        commit_list = self.get_git_log()
        commit_dict = self.sort_commits(commit_list)
        self.build_changelog_file(commit_dict=commit_dict)

        return self.output_file if self.output_file else None

    def get_git_log(self) -> Union[List[str], None]:
        # https://stackoverflow.com/a/12083016/6725620
        # uses only latest tag for this branch
        proc = self.shell_cmd_runner(
            'git log "$(git describe --tags --abbrev=0)..HEAD" --oneline'
        )
        if proc.stdout and proc.stdout != '':
            return proc.stdout.strip().split('\n')
        return None

    def sort_commits(self, commits: List[str]):
        # get the commit types from the toml
        commit_types = self.config.get('commit_types')

        commit_dict = {}
        if commits and len(commits) > 0:
            for commit in commits:
                commit = commit.split(' ', maxsplit=1)
                for commit_type in commit_types:
                    reg = re.compile(
                        fr'{commit_type["message"]}\s?[:|-]', flags=re.I
                    )
                    match = reg.match(commit[1])
                    if match:
                        cleaned_msg = (
                            commit[1].replace(match.group(0), '').strip()
                        )
                        info(
                            f'{cleaned_msg} [{commit[0]}](https://github.com/'
                            f'{self.space}/{self.project}/commit/{commit[0]})'
                        )
                        if commit_type['group'] not in commit_dict:
                            commit_dict[commit_type['group']] = []
                        commit_dict[commit_type['group']].append(
                            f'{cleaned_msg} [{commit[0]}](https://github.com/'
                            f'{self.space}/{self.project}/commit/{commit[0]})'
                        )
        else:
            warning("No conventional commits found.")
            sys.exit(1)
        if not commit_dict:
            warning("No conventional commits found.")
            sys.exit(1)
        return commit_dict

    def build_changelog_file(self, commit_dict: Dict):
        # changelog header
        changelog = (
            "# Changelog\n\n"
            "All notable changes to this project "
            "will be documented in this file.\n\n"
        )
        if self.next_version:
            changelog += (
                f'## [{self.next_version}] - {date.today().isoformat()}\n'
            )
        else:
            changelog += '## [Unreleased]\n\n'
        commit_types = self.config.get('commit_types')

        # changelog entries
        for commit_type in commit_types:
            if commit_type['group'] in commit_dict.keys():
                changelog += f"\n## {commit_type['group']}\n"
                for msg in commit_dict[commit_type['group']]:
                    changelog += f"* {msg}\n"

        # comparison line
        if self.next_version and self.current_version:
            changelog += (
                f'\n[{self.next_version}]: https://github.com/{self.space}/'
                f'{self.project}/compare/{self.current_version}...'
                f'{self.next_version}'
            )
        elif self.current_version:
            changelog += (
                f'\n[Unreleased]: https://github.com/{self.space}/'
                f'{self.project}/compare/{self.current_version}...HEAD'
            )
        else:
            changelog += (
                f'\n[Unreleased]: https://github.com/{self.space}/'
                f'{self.project}/compare/<?>...HEAD'
            )

        self.output_file.write_text(changelog, encoding='utf-8')


def initialize_default_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description='Conventional commits utility.',
        prog='pontos-changelog',
    )

    parser.add_argument(
        '--config',
        '-C',
        default=Path('changelog.toml'),
        type=FileType('r'),
        help="Conventional commits config file (toml), including conventions.",
    )

    parser.add_argument(
        '--project',
        help='The github project',
    )

    parser.add_argument(
        '--space',
        default='greenbone',
        help='User/Team name in github',
    )

    parser.add_argument(
        '--current-version',
        default='greenbone',
        help='Current version before these changes',
    )

    parser.add_argument(
        '--next-version',
        help='The planned release version',
    )

    parser.add_argument(
        '--output',
        '-o',
        default=Path('unreleased.md'),
        type=FileType('r'),
        help='The path to the output file (.md)',
    )

    return parser


def main(
    shell_cmd_runner=lambda x: subprocess.run(
        x,
        shell=True,
        check=True,
        errors="utf-8",  # use utf-8 encoding for error output
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ),
    args=None,
):
    term = Terminal()
    _set_terminal(term)

    term.bold_info('pontos-changelog')

    parser = initialize_default_parser()
    parsed_args = parser.parse_args(args)

    with term.indent():
        try:
            changelog_builder = ChangelogBuilder(
                shell_cmd_runner=shell_cmd_runner,
                args=parsed_args,
            )
            changelog_builder.create_changelog_file()
        except subprocess.CalledProcessError as e:
            error(f'Could not run command "{e.cmd}".')
            out(f'Error was: {e.stderr}')
            sys.exit(1)

    return sys.exit(0)



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from argparse import Namespace
import sys

import requests

from pontos.github.api import GitHubRESTApi
from pontos.terminal import error, ok


def pull_request(args: Namespace):
    git = GitHubRESTApi(token=args.token)

    # check if branches exist
    try:
        if not git.branch_exists(repo=args.repo, branch=args.head):
            error(
                f"Head branch {args.head} is not existing "
                "or authorisation failed."
            )
            sys.exit(1)
        ok(f"Head branch {args.head} is existing.")

        if not git.branch_exists(repo=args.repo, branch=args.target):
            error(
                f"Target branch {args.target} is not existing or "
                "authorisation failed."
            )
            sys.exit(1)
        ok(f"Head branch {args.head} is existing.")

        git.create_pull_request(
            repo=args.repo,
            head_branch=args.head,
            base_branch=args.target,
            title=args.title,
            body=args.body,
        )
    except requests.exceptions.RequestException as e:
        error(str(e))
        sys.exit(1)

    ok("Pull Request created.")



# -*- coding: utf-8 -*-
# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from pathlib import Path
from unittest.mock import MagicMock

from pontos.version.python import PythonVersionCommand
from pontos.version.helper import VersionError


class GetVersionFromPyprojectTomlTestCase(unittest.TestCase):
    def test_pyproject_toml_file_not_exists(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.__str__.return_value = 'pyproject.toml'
        fake_path.exists.return_value = False

        with self.assertRaisesRegex(
            VersionError, 'pyproject.toml file not found'
        ):
            PythonVersionCommand(project_file_path=fake_path)

        fake_path.exists.assert_called_with()

    def test_no_poerty_section(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.__str__.return_value = 'pyproject.toml'
        fake_path.exists.return_value = True
        fake_path.read_text.return_value = (
            '[tool.pontos.version]\nversion-module-file = "foo.py"'
        )

        with self.assertRaisesRegex(
            VersionError, 'Version information not found in pyproject.toml file'
        ):
            cmd = PythonVersionCommand(project_file_path=fake_path)
            cmd._get_version_from_pyproject_toml()  # pylint: disable=protected-access

        fake_path.exists.assert_called_with()
        fake_path.read_text.assert_called_with(encoding='utf-8')

    def test_empty_poetry_section(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.__str__.return_value = 'pyproject.toml'
        fake_path.exists.return_value = True
        fake_path.read_text.return_value = """
        [tool.poetry]
        [tool.pontos.version]\nversion-module-file = "foo.py"
        """

        with self.assertRaisesRegex(
            VersionError, 'Version information not found in pyproject.toml file'
        ):
            cmd = PythonVersionCommand(project_file_path=fake_path)
            cmd._get_version_from_pyproject_toml()  # pylint: disable=protected-access

        fake_path.exists.assert_called_with()
        fake_path.read_text.assert_called_with(encoding='utf-8')

    def test_get_version(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.__str__.return_value = 'pyproject.toml'
        fake_path.exists.return_value = True
        fake_path.read_text.return_value = """
        [tool.poetry]\nversion = "1.2.3"
        [tool.pontos.version]\nversion-module-file = "foo.py"
        """

        cmd = PythonVersionCommand(project_file_path=fake_path)
        # pylint: disable=protected-access
        version = cmd._get_version_from_pyproject_toml()

        self.assertEqual(version, '1.2.3')

        fake_path.exists.assert_called_with()
        fake_path.read_text.assert_called_with(encoding='utf-8')



# -*- coding: utf-8 -*-
# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# pylint: disable=C0413,W0108

import os
import unittest

from dataclasses import dataclass
from pathlib import Path
from unittest.mock import MagicMock

import requests

from pontos.release.helper import version
from pontos import release, changelog


@dataclass
class StdOutput:
    stdout: bytes


class SignTestCase(unittest.TestCase):
    def setUp(self) -> None:
        os.environ['GITHUB_TOKEN'] = 'foo'
        os.environ['GITHUB_USER'] = 'bar'
        self.valid_gh_release_response = (
            '{"zipball_url": "zip", "tarball_url":'
            ' "tar", "upload_url":"upload"}'
        )

    def test_fail_sign_on_invalid_get_response(self):
        fake_path_class = MagicMock(spec=Path)
        fake_requests = MagicMock(spec=requests)
        fake_get = MagicMock(spec=requests.Response).return_value
        fake_get.status_code = 404
        fake_get.text = self.valid_gh_release_response
        fake_requests.get.return_value = fake_get
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')
        args = [
            'sign',
            '--project',
            'foo',
            '--release-version',
            '0.0.1',
        ]

        def runner(_: str):
            return StdOutput('')

        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _requests=fake_requests,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertFalse(released)

    def test_fail_sign_on_upload_fail(self):
        fake_path_class = MagicMock(spec=Path)
        fake_requests = MagicMock(spec=requests)
        fake_get = MagicMock(spec=requests.Response).return_value
        fake_get.status_code = 200
        fake_get.text = self.valid_gh_release_response
        fake_post = MagicMock(spec=requests.Response).return_value
        fake_post.status_code = 500
        fake_post.text = self.valid_gh_release_response
        fake_requests.post.return_value = fake_post
        fake_requests.get.return_value = fake_get
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')
        args = [
            'sign',
            '--project',
            'foo',
            '--release-version',
            '0.0.1',
        ]

        def runner(_: str):
            return StdOutput('')

        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _requests=fake_requests,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertFalse(released)

    def test_successfully_sign(self):
        fake_path_class = MagicMock(spec=Path)
        fake_requests = MagicMock(spec=requests)
        fake_get = MagicMock(spec=requests.Response).return_value
        fake_get.status_code = 200
        fake_get.text = self.valid_gh_release_response
        fake_requests.get.return_value = fake_get
        fake_post = MagicMock(spec=requests.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        fake_requests.post.return_value = fake_post
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')
        args = [
            'sign',
            '--project',
            'bar',
            '--release-version',
            '0.0.1',
        ]

        called = []

        def runner(cmd: str):
            called.append(cmd)
            return StdOutput('')

        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _requests=fake_requests,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertTrue(released)



# Copyright (C) 2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from pontos.github.argparser import parse_args
from pontos.terminal import Terminal, _set_terminal


def main(args=None):
    parsed_args = parse_args(args)

    term = Terminal()
    _set_terminal(term)

    term.bold_info(f'pontos-github => {parsed_args.func.__name__}')

    with term.indent():
        if not parsed_args.token:
            term.error("A Github User Token is required.")
            sys.exit(1)

        parsed_args.func(args=parsed_args)


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess

from pathlib import Path
from typing import List, Optional, Union


class GitError(subprocess.CalledProcessError):
    """
    Error raised while executing a git command
    """

    def __str__(self) -> str:
        cmd = " ".join(self.cmd)
        return (
            f"Git command '{cmd}' returned "
            f"non-zero exit status {str(self.returncode)}"
        )


def _exec_git(
    *args: str, ignore_errors: Optional[bool] = False, cwd: Optional[str] = None
) -> str:
    """
    Internal module function to abstract calling git via subprocess
    """
    try:
        cmd_args = ["git"]
        cmd_args.extend(args)
        output = subprocess.check_output(cmd_args, cwd=cwd)
        return output.decode()
    except subprocess.CalledProcessError as e:
        if ignore_errors:
            return ""
        raise GitError(e.returncode, e.cmd, e.output, e.stderr) from None


class Git:
    """
    Run git commands as subprocesses
    """

    def __init__(self, cwd: Optional[Path] = None) -> None:
        """
        Create a new Git instance

        Args:
            cwd: Set the current working directory for the git commands
        """
        self._cwd = cwd.absolute() if cwd else None

    @property
    def cwd(self) -> Path:
        """
        Get the current working directory as Path
        """
        return self._cwd

    @cwd.setter
    def cwd(self, cwd: Path):
        """
        Set the current working directory for all following git commands
        """
        self._cwd = cwd.absolute()

    def init(self, *, bare: Optional[bool] = False):
        """
        Init a git repository

        Args:
            bare: Wether to create a `bare` repository or not.
                  Defaults to false.
        """
        args = ["init"]
        if bare:
            args.append("--bare")
        _exec_git(*args, cwd=self._cwd)

    def create_branch(self, branch: str, *, start_point: Optional[str] = None):
        """
        Create a new branch

        Args:
            branch: Name of the branch to be created
            start_point: An optional git reference (branch, tag, sha, ...) from
                         where to start the branch
        """
        args = ["checkout", "-b", branch]
        if start_point:
            args.append(start_point)

        _exec_git(*args, cwd=self._cwd)

    def rebase(
        self,
        base: str,
        *,
        head: Optional[str] = None,
        onto: Optional[str] = None,
    ):
        """
        Rebase a branch

        Args:
            base: Apply changes of this branch
            head: Apply changes on this branch. If not set the current branch is
                  used.
            onto: Apply changes on top of this branch
        """
        args = ["rebase"]

        if onto:
            args.extend(["--onto", onto])

        args.append(base)

        if head:
            args.append(head)

        _exec_git(*args, cwd=self._cwd)

    def clone(
        self,
        repo_url: str,
        destination: Path,
        *,
        branch: Optional[str] = None,
        remote: Optional[str] = None,
        depth: Optional[int] = None,
    ):
        """
        Clone a repository

        Args:
            repo_url: URL of the repo to clone
            destination: Where to checkout the clone
            branch: Branch to checkout. By default the default branch is used.
            remote: Store repo url under this remote name
        """
        args = ["clone"]
        if remote:
            args.extend(["-o", remote])
        if branch:
            args.extend(["-b", branch])
        if depth:
            args.extend(["--depth", depth])
        args.extend([repo_url, str(destination.absolute())])

        _exec_git(
            *args,
            cwd=self._cwd,
        )

    def push(
        self,
        *,
        remote: Optional[str] = None,
        branch: Optional[str] = None,
        follow_tags: bool = False,
    ):
        """
        Push changes to remote repository

        Args:
            remote: Push changes to the named remote
            branch: Branch to push. Will only be considered in combination with
                    a remote.
            follow_tags: Push all tags pointing to a commit included in the to
                         be pushed branch.
        """
        args = ["push"]
        if follow_tags:
            args.append("--follow-tags")
        if remote:
            args.append(remote)
            if branch:
                args.append(branch)

        _exec_git(*args, cwd=self._cwd)

    def config(self, key: str, value: str):
        """
        Set a (local) git config
        """
        _exec_git("config", key, value, cwd=self._cwd)

    def cherry_pick(self, commits: Union[str, List[str]]):
        """
        Apply changes of a commit(s) to the current branch

        Args:
            commit: A single git reference (e.g. sha) of the commit or a list
                    of git references.
        """
        if isinstance(commits, str):
            commits = [commits]

        args = ["cherry-pick"]
        args.extend(commits)

        _exec_git(*args, cwd=self._cwd)

    def list_tags(self) -> List[str]:
        """
        List all available tags
        """
        return _exec_git("tag", "-l", cwd=self._cwd).splitlines()

    def add(self, files: Union[str, List[str]]):
        """
        Add files to the git staging area

        Args:
            files: A single file or a list of files to add to the staging area
        """
        if isinstance(files, str):
            files = [files]

        args = ["add"]
        args.extend(files)

        _exec_git(*args, cwd=self._cwd)

    def commit(
        self,
        message: str,
        *,
        verify: Optional[bool] = None,
        gpg_signing_key: Optional[str] = None,
    ):
        """
        Create a new commit

        Args:
            message: Message of the commit
            verify: Set to False to skip git hooks
            gpg_signing_key: GPG Key ID to use to sign the commit
        """
        args = ["commit"]
        if verify is False:
            args.append("--no-verify")
        if gpg_signing_key:
            args.append(f"-S{gpg_signing_key}")

        args.extend(["-m", message])

        _exec_git(*args, cwd=self._cwd)



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib

from pathlib import Path

import tomlkit

from .helper import (
    safe_version,
    strip_version,
    check_develop,
    is_version_pep440_compliant,
    VersionError,
    versions_equal,
)
from .version import VersionCommand


TEMPLATE = """# pylint: disable=invalid-name

# THIS IS AN AUTOGENERATED FILE. DO NOT TOUCH!

__version__ = "{}"\n"""


# This class is used for Python Version command(s)
class PythonVersionCommand(VersionCommand):
    def __init__(self, *, project_file_path: Path = None) -> None:
        if not project_file_path:
            project_file_path = Path.cwd() / 'pyproject.toml'

        if not project_file_path.exists():
            raise VersionError(f'{str(project_file_path)} file not found.')

        self.pyproject_toml = tomlkit.parse(
            project_file_path.read_text(encoding='utf-8')
        )

        if (
            'tool' not in self.pyproject_toml
            or 'pontos' not in self.pyproject_toml['tool']
            or 'version' not in self.pyproject_toml['tool']['pontos']
        ):
            raise VersionError(
                '[tool.pontos.version] section missing '
                f'in {str(project_file_path)}.'
            )

        pontos_version_settings = self.pyproject_toml['tool']['pontos'][
            'version'
        ]

        try:
            version_file_path = Path(
                pontos_version_settings['version-module-file']
            )
        except tomlkit.exceptions.NonExistentKey:
            raise VersionError(
                'version-module-file key not set in [tool.pontos.version] '
                f'section of {str(project_file_path)}.'
            ) from None

        super().__init__(
            version_file_path=version_file_path,
            project_file_path=project_file_path,
        )

    def _get_version_from_pyproject_toml(self) -> str:
        """
        Return the version information from the [tool.poetry] section of the
        pyproject.toml file. The version may be in non standardized form.
        """

        if (
            'tool' in self.pyproject_toml
            and 'poetry' in self.pyproject_toml['tool']
            and 'version' in self.pyproject_toml['tool']['poetry']
        ):
            return self.pyproject_toml['tool']['poetry']['version']

        raise VersionError(
            'Version information not found in '
            f'{str(self.project_file_path)} file.'
        )

    def _update_version_file(self, new_version: str) -> None:
        """
        Update the version file with the new version
        """
        new_version = safe_version(new_version)
        self.version_file_path.write_text(
            TEMPLATE.format(new_version), encoding='utf-8'
        )

    def _update_pyproject_version(
        self,
        new_version: str,
    ) -> None:
        """
        Update the version in the pyproject.toml file
        """

        new_version = safe_version(new_version)
        pyproject_toml = tomlkit.parse(
            self.project_file_path.read_text(encoding='utf-8')
        )

        if 'tool' not in pyproject_toml:
            tool_table = tomlkit.table()
            pyproject_toml['tool'] = tool_table

        if 'poetry' not in pyproject_toml['tool']:
            poetry_table = tomlkit.table()
            pyproject_toml['tool'].add('poetry', poetry_table)

        pyproject_toml['tool']['poetry']['version'] = new_version

        self.project_file_path.write_text(
            tomlkit.dumps(pyproject_toml), encoding='utf-8'
        )

    def get_current_version(self) -> str:
        version_module_name = self.version_file_path.stem
        module_parts = list(self.version_file_path.parts[:-1]) + [
            version_module_name
        ]
        module_name = '.'.join(module_parts)
        try:
            version_module = importlib.import_module(module_name)
        except ModuleNotFoundError:
            raise VersionError(
                f'Could not load version from {module_name}. Import failed.'
            ) from None

        return version_module.__version__

    def verify_version(self, version: str) -> None:
        current_version = self.get_current_version()
        if not is_version_pep440_compliant(current_version):
            raise VersionError(
                f"The version {current_version} in "
                f"{str(self.version_file_path)} is not PEP 440 compliant."
            )

        pyproject_version = self._get_version_from_pyproject_toml()

        if pyproject_version != current_version:
            raise VersionError(
                f"The version {pyproject_version} in "
                f"{str(self.project_file_path)} doesn't match the current "
                f"version {current_version}."
            )

        if version != 'current':
            provided_version = strip_version(version)
            if provided_version != current_version:
                raise VersionError(
                    f"Provided version {provided_version} does not match the "
                    f"current version {current_version}."
                )

        self._print('OK')

    def update_version(
        self, new_version: str, *, develop: bool = False, force: bool = False
    ) -> None:

        new_version = safe_version(new_version)
        if check_develop(new_version) and develop:
            develop = False
        if develop:
            new_version = f'{new_version}.dev1'

        pyproject_version = self._get_version_from_pyproject_toml()

        if not self.version_file_path.exists():
            self.version_file_path.touch()
        elif not force and versions_equal(
            new_version, self.get_current_version()
        ):
            self._print('Version is already up-to-date.')
            return

        self._update_pyproject_version(new_version=new_version)

        self._update_version_file(new_version=new_version)

        self._print(
            f'Updated version from {pyproject_version} to {new_version}'
        )



# -*- coding: utf-8 -*-
# pontos/release/release.py
# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from argparse import Namespace
import sys

from pathlib import Path
from typing import Callable

from pontos import changelog
from pontos.terminal import error, warning, info, ok, out
from pontos import version

from .helper import (
    calculate_calendar_version,
    commit_files,
    find_signing_key,
    get_project_name,
    get_current_version,
    get_next_patch_version,
    update_version,
)

RELEASE_TEXT_FILE = ".release.md"


def prepare(
    shell_cmd_runner: Callable,
    args: Namespace,
    *,
    path: Path,
    version_module: version,
    changelog_module: changelog,
    **_kwargs,
) -> bool:
    git_tag_prefix: str = args.git_tag_prefix
    git_signing_key: str = (
        args.git_signing_key
        if args.git_signing_key is not None
        else find_signing_key(shell_cmd_runner)
    )
    project: str = (
        args.project
        if args.project is not None
        else get_project_name(shell_cmd_runner)
    )
    space: str = args.space
    calendar: bool = args.calendar
    patch: bool = args.patch

    if calendar:
        release_version: str = calculate_calendar_version()
    elif patch:
        release_version: str = get_next_patch_version()
    else:
        release_version: str = args.release_version

    info(f"Preparing the release {release_version}")

    # guardian
    git_tags = shell_cmd_runner('git tag -l')
    git_version = f"{git_tag_prefix}{release_version}"
    if git_version.encode() in git_tags.stdout.splitlines():
        error(f'git tag {git_version} is already taken.')
        sys.exit(1)

    executed, filename = update_version(release_version, version_module)
    if not executed:
        return False

    ok(f"updated version  in {filename} to {release_version}")

    changelog_bool = True
    if args.conventional_commits:
        current_version = get_current_version()
        output = f'v{release_version}.md'
        cargs = Namespace(
            current_version=current_version,
            next_version=release_version,
            output=output,
            space=space,
            project=project,
            config=args.cc_config,
        )
        changelog_builder = changelog_module.ChangelogBuilder(
            shell_cmd_runner=shell_cmd_runner,
            args=cargs,
        )

        output_file = changelog_builder.create_changelog_file()
        ok(f"Created changelog {output}")
        commit_msg = f'Changelog created for release to {release_version}'
        commit_files(
            output_file,
            commit_msg,
            shell_cmd_runner,
            git_signing_key=git_signing_key,
        )
        changelog_bool = False
        # Remove the header for the release text
        changelog_text = output_file.read_text(encoding='utf-8').replace(
            "# Changelog\n\n"
            "All notable changes to this project "
            "will be documented in this file.\n\n",
            '',
        )
    else:
        change_log_path = path.cwd() / 'CHANGELOG.md'
        if args.changelog:
            tmp_path = path.cwd() / Path(args.changelog)
            if tmp_path.is_file():
                change_log_path = tmp_path
            else:
                warning(f"{tmp_path} is not a file.")

        # Try to get the unreleased section of the specific version
        updated, changelog_text = changelog_module.update(
            change_log_path.read_text(encoding='utf-8'),
            release_version,
            git_tag_prefix=git_tag_prefix,
            containing_version=release_version,
        )

        if not updated:
            # Try to get unversioned unrelease section
            updated, changelog_text = changelog_module.update(
                change_log_path.read_text(encoding='utf-8'),
                release_version,
                git_tag_prefix=git_tag_prefix,
            )

        if not updated:
            error("No unreleased text found in CHANGELOG.md")
            sys.exit(1)

        change_log_path.write_text(updated, encoding='utf-8')

        ok("Updated CHANGELOG.md")

    info("Committing changes")

    commit_msg = f'Automatic release to {release_version}'
    commit_files(
        filename,
        commit_msg,
        shell_cmd_runner,
        git_signing_key=git_signing_key,
        changelog=changelog_bool,
    )

    if git_signing_key:
        shell_cmd_runner(
            f"git tag -u {git_signing_key} {git_version} -m '{commit_msg}'"
        )
    else:
        shell_cmd_runner(f"git tag {git_version} -m '{commit_msg}'")

    release_text = path(RELEASE_TEXT_FILE)
    release_text.write_text(changelog_text, encoding='utf-8')

    warning(
        f"Please verify git tag {git_version}, "
        f"commit and release text in {str(release_text)}"
    )
    out("Afterwards please execute release")

    return True



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
from typing import Callable, Dict, Iterator, Optional

import requests

DEFAULT_GITHUB_API_URL = "https://api.github.com"

DEFAULT_TIMEOUT = 1000
DEFAULT_CHUNK_SIZE = 4096


class DownloadProgressIterable:
    def __init__(
        self, content_iterator: Iterator, destination: Path, length: int
    ):
        self._length = None if length is None else int(length)
        self._content_iterator = content_iterator
        self._destination = destination

    @property
    def length(self) -> Optional[int]:
        """
        Size in bytes of the to be downloaded file or None if the size is not
        available
        """
        return self._length

    @property
    def destination(self) -> Path:
        """
        Destination path of the to be downloaded file
        """
        return self._destination

    def _download(self) -> Iterator[Optional[float]]:
        dl = 0
        with self._destination.open("wb") as f:
            for content in self._content_iterator:
                dl += len(content)
                f.write(content)
                yield dl / self._length if self._length else None

    def __iter__(self) -> Iterator[Optional[float]]:
        return self._download()

    def run(self):
        """
        Just run the download without caring about the progress
        """
        try:
            it = iter(self)
            while True:
                next(it)
        except StopIteration:
            pass


def download(
    url: str,
    destination: Optional[Path] = None,
    *,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    timeout: int = DEFAULT_TIMEOUT,
) -> DownloadProgressIterable:
    """Download file in url to filename

    Arguments:
        url: The url of the file we want to download
        destination: Path of the file to store the download in. If set it will
                     be derived from the passed URL.
        chunk_size: Download file in chunks of this size
        timeout: Connection timeout

    Raises:
        HTTPError if the request was invalid

    Returns:
        A DownloadProgressIterator that yields the progress of the download in
        percent for each downloaded chunk or None for each chunk if the progress
        is unknown.
    """
    destination = Path(url.split('/')[-1]) if not destination else destination

    response = requests.get(url, stream=True, timeout=timeout)
    response.raise_for_status()

    total_length = response.headers.get('content-length')

    return DownloadProgressIterable(
        response.iter_content(chunk_size=chunk_size),
        destination,
        total_length,
    )


class GitHubRESTApi:
    def __init__(
        self, token: str, url: Optional[str] = DEFAULT_GITHUB_API_URL
    ) -> None:
        self.token = token
        self.url = url

    def _request(
        self,
        api: str,
        *,
        params: Optional[Dict[str, str]] = None,
        data: Optional[Dict[str, str]] = None,
        request: Optional[Callable] = None,
    ) -> requests.Response:
        headers = {
            "Authorization": f"token {self.token}",
            "Accept": "application/vnd.github.v3+json",
        }
        request = request or requests.get
        return request(
            f"{self.url}{api}", headers=headers, params=params, json=data
        )

    def branch_exists(self, repo: str, branch: str) -> bool:
        """
        Check if a single branch in a repository exists

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Branch name to check
        """
        api = f"/repos/{repo}/branches/{branch}"
        response = self._request(api)
        return response.ok

    def pull_request_commits(
        self, repo: str, pull_request: str
    ) -> Dict[str, str]:
        """
        Get all commit information of a pull request

        Hint: At maximum GitHub allows to receive 100 commits. If a pull request
        contains more then 100 commits only the first 100 are returned.

        Args:
            repo: GitHub repository (owner/name) to use
            pull_request: Pull request number

        Returns:
            Information about the commits in the pull request as a dict
        """
        # per default github only shows 35 commits and at max it is only
        # possible to receive 100
        params = {"per_page": "100"}
        api = f"/repos/{repo}/pulls/{pull_request}/commits"
        response = self._request(api, params=params)
        return response.json()

    def create_pull_request(
        self,
        repo: str,
        *,
        head_branch: str,
        base_branch: str,
        title: str,
        body: str,
    ):
        """
        Create a new Pull Request on GitHub

        Args:
            repo: GitHub repository (owner/name) to use
            head_branch: Branch to create a pull request from
            base_branch: Branch as target for the pull
            title: Title for the pull request
            body: Description for the pull request. Can be formatted in Markdown

        Raises:
            HTTPError if the request was invalid
        """
        api = f"/repos/{repo}/pulls"
        data = {
            "head": head_branch,
            "base": base_branch,
            "title": title,
            "body": body,
        }
        response = self._request(api, data=data, request=requests.post)
        response.raise_for_status()

    def add_pull_request_comment(
        self, repo: str, pull_request: str, comment: str
    ):
        """
        Add a comment to a pull request on GitHub

        Args:
            repo: GitHub repository (owner/name) to use
            pull_request: Pull request number where to add a comment
            comment: The actual comment message. Can be formatted in Markdown.

        Raises:
            HTTPError if the request was invalid
        """
        api = f"/repos/{repo}/issues/{pull_request}/comments"
        data = {"body": comment}
        response = self._request(api, data=data, request=requests.post)
        response.raise_for_status()

    def delete_branch(self, repo: str, branch: str):
        """
        Delete a branch on GitHub

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Branch to be deleted

        Raises:
            HTTPError if the request was invalid
        """
        api = f"/repos/{repo}/git/refs/{branch}"
        response = self._request(api, request=requests.delete)
        response.raise_for_status()

    def create_release(
        self,
        repo: str,
        tag: str,
        *,
        body: Optional[str] = None,
        name: Optional[str] = None,
        target_commitish: Optional[str] = None,
        draft: Optional[bool] = False,
        prerelease: Optional[bool] = False,
    ):
        """
        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            body: Content of the changelog for the release
            name: name of the release, e.g. 'pontos 1.0.0'
            target_commitish: Only needed when tag is not there yet
            draft: If the release is a draft. False by default.
            prerelease: If the release is a pre release. False by default.
        """
        data = {
            'tag_name': tag,
            'target_commitish': target_commitish,
            'name': name,
            'body': body,
            'draft': draft,
            'prerelease': prerelease,
        }
        api = f"/repos/{repo}/releases"
        response = self._request(api, data=data, request=requests.post)
        response.raise_for_status()

    def release_exists(self, repo: str, tag: str) -> bool:
        """
        Check wether a GitHub release exists by tag

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Returns:
            True if the release exists
        """
        api = f"/repos/{repo}/releases/tags/{tag}"
        response = self._request(api)
        return response.ok

    def release(self, repo: str, tag: str) -> Dict[str, str]:
        """
        Get data of a GitHub release by tag

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Raises:
            HTTPError if the request was invalid
        """
        api = f"/repos/{repo}/releases/tags/{tag}"
        response = self._request(api)
        response.raise_for_status()
        return response.json()

    def download_release_tarball(
        self, repo: str, tag: str, destination: Path
    ) -> DownloadProgressIterable:
        """
        Download a release tarball (tar.gz) file

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            destination: A path where to store the downloaded file

        Raises:
            HTTPError if the request was invalid
        """
        api = f"https://github.com/{repo}/archive/refs/tags/{tag}.tar.gz"
        return download(api, destination)

    def download_release_zip(
        self, repo: str, tag: str, destination: Path
    ) -> DownloadProgressIterable:
        """
        Download a release zip file

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            destination: A path where to store the downloaded file

        Raises:
            HTTPError if the request was invalid
        """
        api = f"https://github.com/{repo}/archive/refs/tags/{tag}.zip"
        return download(api, destination)



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from argparse import Namespace
from pathlib import Path
import unittest
from unittest.mock import Mock

from pontos.github.argparser import parse_args
from pontos.github.cmds import pull_request


class TestArgparsing(unittest.TestCase):
    def setUp(self):
        self.term = Mock()

    def test_parse_args(self):
        argv = [
            "pr",
            "foo/bar",
            "baz",
            "main",
            "baz in main",
        ]

        parsed_args = parse_args(argv)

        template = Path().cwd() / 'pontos/github/pr_template.md'

        expected_args = Namespace(
            body=template.read_text(encoding='utf-8'),
            command='pr',
            func=pull_request,
            head='baz',
            repo='foo/bar',
            target='main',
            title='baz in main',
            token='GITHUB_TOKEN',
        )

        self.assertEqual(parsed_args, expected_args)

    def test_parse_args_fail(self):
        argv = ["pr", "foo/bar"]

        with self.assertRaises(SystemExit):
            parse_args(argv)



# -*- coding: utf-8 -*-
# pontos/release/release.py
# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from argparse import Namespace
import json

from pathlib import Path
from typing import Callable

import requests

from pontos import changelog
from pontos.terminal import error, warning, info
from pontos import version

from .helper import (
    build_release_dict,
    commit_files,
    find_signing_key,
    get_current_version,
    get_next_dev_version,
    get_project_name,
    update_version,
)

RELEASE_TEXT_FILE = ".release.md"


def release(
    shell_cmd_runner: Callable,
    args: Namespace,
    *,
    path: Path,
    version_module: version,
    username: str,
    token: str,
    requests_module: requests,
    changelog_module: changelog,
    **_kwargs,
) -> bool:
    project: str = (
        args.project
        if args.project is not None
        else get_project_name(shell_cmd_runner)
    )
    space: str = args.space
    git_signing_key: str = (
        args.git_signing_key
        if args.git_signing_key is not None
        else find_signing_key(shell_cmd_runner)
    )
    git_remote_name: str = (
        args.git_remote_name if args.git_remote_name is not None else ''
    )
    git_tag_prefix: str = args.git_tag_prefix
    release_version: str = (
        args.release_version
        if args.release_version is not None
        else get_current_version()
    )
    next_version: str = (
        args.next_version
        if args.next_version is not None
        else get_next_dev_version(release_version)
    )

    info("Pushing changes")

    shell_cmd_runner(f"git push --follow-tags {git_remote_name}")

    info(f"Creating release for v{release_version}")
    changelog_text: str = path(RELEASE_TEXT_FILE).read_text(encoding='utf-8')

    headers = {'Accept': 'application/vnd.github.v3+json'}

    base_url = f"https://api.github.com/repos/{space}/{project}/releases"
    git_version = f'{git_tag_prefix}{release_version}'
    response = requests_module.post(
        base_url,
        headers=headers,
        auth=(username, token),
        json=build_release_dict(
            git_version,
            changelog_text,
            name=f"{project} {release_version}",
        ),
    )
    if response.status_code != 201:
        error(f"Wrong response status code: {response.status_code}")
        error(json.dumps(response.text, indent=4, sort_keys=True))
        return False

    path(RELEASE_TEXT_FILE).unlink()

    commit_msg = (
        f'Automatic adjustments after release\n\n'
        f'* Update to version {next_version}\n'
    )

    # set to new version add skeleton
    changelog_bool = True
    if not args.conventional_commits:
        change_log_path = path.cwd() / 'CHANGELOG.md'
        if args.changelog:
            tmp_path = path.cwd() / Path(args.changelog)
            if tmp_path.is_file():
                change_log_path = tmp_path
            else:
                warning(f"{tmp_path} is not a file.")

        updated = changelog_module.add_skeleton(
            markdown=change_log_path.read_text(encoding='utf-8'),
            new_version=release_version,
            project_name=project,
            git_tag_prefix=git_tag_prefix,
            git_space=space,
        )
        change_log_path.write_text(updated, encoding='utf-8')
        changelog_bool = False

        commit_msg += f'* Add empty changelog after {release_version}'

    executed, filename = update_version(
        next_version, version_module, develop=True
    )
    if not executed:
        return False

    commit_files(
        filename,
        commit_msg,
        shell_cmd_runner,
        git_signing_key=git_signing_key,
        changelog=changelog_bool,
    )

    # pushing the new tag
    shell_cmd_runner(f"git push --follow-tags {git_remote_name}")

    return True



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import unittest

from pathlib import Path
from unittest.mock import patch, MagicMock

from pontos.github.api import GitHubRESTApi, DEFAULT_TIMEOUT, download

here = Path(__file__).parent


class GitHubApiTestCase(unittest.TestCase):
    @patch("pontos.github.api.requests.get")
    def test_branch_exists(self, requests_mock: MagicMock):
        response = MagicMock()
        response.ok = True
        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        exists = api.branch_exists("foo/bar", "main")

        requests_mock.assert_called_once_with(
            'https://api.github.com/repos/foo/bar/branches/main',
            headers={
                'Authorization': 'token 12345',
                'Accept': 'application/vnd.github.v3+json',
            },
            params=None,
            json=None,
        )
        self.assertTrue(exists)

    @patch("pontos.github.api.requests.get")
    def test_branch_not_exists(self, requests_mock: MagicMock):
        response = MagicMock()
        response.ok = False
        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        exists = api.branch_exists("foo/bar", "main")

        requests_mock.assert_called_once_with(
            'https://api.github.com/repos/foo/bar/branches/main',
            headers={
                'Authorization': 'token 12345',
                'Accept': 'application/vnd.github.v3+json',
            },
            params=None,
            json=None,
        )
        self.assertFalse(exists)

    @patch("pontos.github.api.requests.get")
    def test_pull_request_commits(self, requests_mock: MagicMock):
        response = MagicMock()
        response.json.return_value = [{"sha": "1"}]
        requests_mock.return_value = response
        api = GitHubRESTApi("12345")
        commits = api.pull_request_commits("foo/bar", "1")

        requests_mock.assert_called_once_with(
            'https://api.github.com/repos/foo/bar/pulls/1/commits',
            headers={
                'Authorization': 'token 12345',
                'Accept': 'application/vnd.github.v3+json',
            },
            params={'per_page': '100'},
            json=None,
        )

        self.assertEqual(len(commits), 1)
        self.assertEqual(commits[0]["sha"], "1")

    @patch("pontos.github.api.requests.post")
    def test_create_pull_request(self, requests_mock: MagicMock):
        api = GitHubRESTApi("12345")
        api.create_pull_request(
            "foo/bar",
            head_branch="foo",
            base_branch="main",
            title="Foo",
            body="This is bar",
        )

        requests_mock.assert_called_once_with(
            'https://api.github.com/repos/foo/bar/pulls',
            headers={
                'Authorization': 'token 12345',
                'Accept': 'application/vnd.github.v3+json',
            },
            params=None,
            json={
                'head': 'foo',
                'base': 'main',
                'title': 'Foo',
                'body': 'This is bar',
            },
        )

    @patch("pontos.github.api.requests.post")
    def test_add_pull_request_comment(self, requests_mock: MagicMock):
        api = GitHubRESTApi("12345")
        api.add_pull_request_comment("foo/bar", "123", "This is a comment")

        requests_mock.assert_called_once_with(
            'https://api.github.com/repos/foo/bar/issues/123/comments',
            headers={
                'Authorization': 'token 12345',
                'Accept': 'application/vnd.github.v3+json',
            },
            params=None,
            json={'body': 'This is a comment'},
        )

    @patch("pontos.github.api.requests.delete")
    def test_delete_branch(self, requests_mock: MagicMock):
        api = GitHubRESTApi("12345")
        api.delete_branch("foo/bar", "foo")

        requests_mock.assert_called_once_with(
            'https://api.github.com/repos/foo/bar/git/refs/foo',
            headers={
                'Authorization': 'token 12345',
                'Accept': 'application/vnd.github.v3+json',
            },
            params=None,
            json=None,
        )

    @patch("pontos.github.api.requests.post")
    def test_create_release(self, requests_mock: MagicMock):
        api = GitHubRESTApi("12345")
        api.create_release(
            "foo/bar",
            "v1.2.3",
            name="Foo v1.2.3",
            body="This is a release",
        )

        requests_mock.assert_called_once_with(
            'https://api.github.com/repos/foo/bar/releases',
            headers={
                'Authorization': 'token 12345',
                'Accept': 'application/vnd.github.v3+json',
            },
            params=None,
            json={
                'tag_name': 'v1.2.3',
                'target_commitish': None,
                'name': 'Foo v1.2.3',
                'body': 'This is a release',
                'draft': False,
                'prerelease': False,
            },
        )

    @patch("pontos.github.api.requests.get")
    def test_release_exists(self, requests_mock: MagicMock):
        response = MagicMock()
        response.ok = True
        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        exists = api.release_exists("foo/bar", "v1.2.3")

        requests_mock.assert_called_once_with(
            'https://api.github.com/repos/foo/bar/releases/tags/v1.2.3',
            headers={
                'Authorization': 'token 12345',
                'Accept': 'application/vnd.github.v3+json',
            },
            params=None,
            json=None,
        )
        self.assertTrue(exists)

    @patch("pontos.github.api.requests.get")
    def test_release(self, requests_mock: MagicMock):
        response = MagicMock()
        response.json.return_value = json.loads(
            (here / "release-response.json").read_text(encoding="utf-8")
        )

        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        data = api.release("greenbone/pontos", "v21.11.0")

        requests_mock.assert_called_once_with(
            'https://api.github.com/repos/greenbone/pontos/releases/tags/v21.11.0',  # pylint: disable=line-too-long
            headers={
                'Authorization': 'token 12345',
                'Accept': 'application/vnd.github.v3+json',
            },
            params=None,
            json=None,
        )

        self.assertEqual(data["id"], 52499047)

    @patch("pontos.github.api.Path")
    @patch("pontos.github.api.requests.get")
    def test_download_release_tarball(
        self, requests_mock: MagicMock, path_mock: MagicMock
    ):
        response = MagicMock()
        response.iter_content.return_value = ["foo", "bar", "baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = None
        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        download_file = path_mock()
        download_progress = api.download_release_tarball(
            "greenbone/pontos", "v21.11.0", download_file
        )

        requests_mock.assert_called_once_with(
            'https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz',  # pylint: disable=line-too-long
            stream=True,
            timeout=DEFAULT_TIMEOUT,
        )
        response_headers.get.assert_called_once_with('content-length')

        self.assertIsNone(download_progress.length)

        it = iter(download_progress)
        progress = next(it)
        self.assertIsNone(progress)
        progress = next(it)
        self.assertIsNone(progress)
        progress = next(it)
        self.assertIsNone(progress)

        with self.assertRaises(StopIteration):
            next(it)

    @patch("pontos.github.api.Path")
    @patch("pontos.github.api.requests.get")
    def test_download_release_tarball_with_content_length(
        self, requests_mock: MagicMock, path_mock: MagicMock
    ):
        response = MagicMock()
        response.iter_content.return_value = ["foo", "bar", "baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = "9"
        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        download_file = path_mock()
        download_progress = api.download_release_tarball(
            "greenbone/pontos", "v21.11.0", download_file
        )

        requests_mock.assert_called_once_with(
            'https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz',  # pylint: disable=line-too-long
            stream=True,
            timeout=DEFAULT_TIMEOUT,
        )
        response_headers.get.assert_called_once_with('content-length')

        self.assertEqual(download_progress.length, 9)

        it = iter(download_progress)
        progress = next(it)
        self.assertEqual(progress, 1 / 3)
        progress = next(it)
        self.assertEqual(progress, 2 / 3)
        progress = next(it)
        self.assertEqual(progress, 1)

        with self.assertRaises(StopIteration):
            next(it)

    @patch("pontos.github.api.Path")
    @patch("pontos.github.api.requests.get")
    def test_download_release_zip(
        self, requests_mock: MagicMock, path_mock: MagicMock
    ):
        response = MagicMock()
        response.iter_content.return_value = ["foo", "bar", "baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = None
        requests_mock.return_value = response

        api = GitHubRESTApi("12345")
        download_file = path_mock()
        download_progress = api.download_release_zip(
            "greenbone/pontos", "v21.11.0", download_file
        )

        requests_mock.assert_called_once_with(
            'https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.zip',  # pylint: disable=line-too-long
            stream=True,
            timeout=DEFAULT_TIMEOUT,
        )
        response_headers.get.assert_called_once_with('content-length')

        self.assertIsNone(download_progress.length)

        it = iter(download_progress)
        progress = next(it)
        self.assertIsNone(progress)
        progress = next(it)
        self.assertIsNone(progress)
        progress = next(it)
        self.assertIsNone(progress)

        with self.assertRaises(StopIteration):
            next(it)


class DownloadTestCase(unittest.TestCase):
    @patch("pontos.github.api.requests.get")
    def test_download_without_destination(
        self,
        requests_mock: MagicMock,
    ):
        response = MagicMock()
        response.iter_content.return_value = [b"foo", b"bar", b"baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = None
        requests_mock.return_value = response

        download_progress = download(
            "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz"  # pylint: disable=line-too-long
        )

        requests_mock.assert_called_once_with(
            'https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz',  # pylint: disable=line-too-long
            stream=True,
            timeout=DEFAULT_TIMEOUT,
        )
        response_headers.get.assert_called_once_with('content-length')

        self.assertIsNone(download_progress.length)
        self.assertEqual(download_progress.destination, Path("v21.11.0.tar.gz"))

        it = iter(download_progress)
        progress = next(it)
        self.assertIsNone(progress)
        progress = next(it)
        self.assertIsNone(progress)
        progress = next(it)
        self.assertIsNone(progress)

        with self.assertRaises(StopIteration):
            next(it)

        download_progress.destination.unlink()

    @patch("pontos.github.api.Path")
    @patch("pontos.github.api.requests.get")
    def test_download_with_content_length(
        self, requests_mock: MagicMock, path_mock: MagicMock
    ):
        response = MagicMock()
        response.iter_content.return_value = ["foo", "bar", "baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = "9"
        requests_mock.return_value = response

        download_file = path_mock()
        file_mock = MagicMock()
        file_mock.__enter__.return_value = file_mock
        download_file.open.return_value = file_mock

        download_progress = download(
            "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz",  # pylint: disable=line-too-long
            download_file,
        )

        requests_mock.assert_called_once_with(
            "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz",  # pylint: disable=line-too-long
            stream=True,
            timeout=DEFAULT_TIMEOUT,
        )
        response_headers.get.assert_called_once_with('content-length')

        self.assertEqual(download_progress.length, 9)

        it = iter(download_progress)

        progress = next(it)
        self.assertEqual(progress, 1 / 3)
        file_mock.write.assert_called_with("foo")

        progress = next(it)
        self.assertEqual(progress, 2 / 3)
        file_mock.write.assert_called_with("bar")

        progress = next(it)
        self.assertEqual(progress, 1)
        file_mock.write.assert_called_with("baz")

        with self.assertRaises(StopIteration):
            next(it)



# -*- coding: utf-8 -*-
# pontos/release/release.py
# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from argparse import FileType, Namespace, ArgumentParser
import sys
import subprocess
import os

from pathlib import Path
from typing import Tuple

import requests

from pontos import changelog
from pontos.terminal import _set_terminal, error, out
from pontos.terminal.terminal import Terminal
from pontos import version

from .prepare import prepare
from .sign import sign
from .release import release


def initialize_default_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description='Release handling utility.',
        prog='pontos-release',
    )

    subparsers = parser.add_subparsers(
        title='subcommands',
        description='valid subcommands',
        help='additional help',
        dest='command',
    )

    prepare_parser = subparsers.add_parser('prepare')
    prepare_parser.set_defaults(func=prepare)
    version_group = prepare_parser.add_mutually_exclusive_group(required=True)
    version_group.add_argument(
        '--release-version',
        help='Will release changelog as version. Must be PEP 440 compliant',
    )
    version_group.add_argument(
        '--calendar',
        help=(
            'Automatically calculate calendar release version, from current'
            ' version and date.'
        ),
        action='store_true',
    )
    version_group.add_argument(
        '--patch',
        help=('Release next patch version: ' 'e.g. x.x.3 -> x.x.4'),
        action='store_true',
    )

    prepare_parser.add_argument(
        '--git-signing-key',
        help='The key to sign the commits and tag for a release',
    )
    prepare_parser.add_argument(
        '--git-tag-prefix',
        default='v',
        help='Prefix for git tag versions. Default: %(default)s',
    )
    prepare_parser.add_argument(
        '--changelog',
        help=(
            'The CHANGELOG file path, defaults '
            'to CHANGELOG.md in the repository root directory'
        ),
    )
    prepare_parser.add_argument(
        '--space',
        default='greenbone',
        help='User/Team name in github',
    )
    prepare_parser.add_argument(
        '--project',
        help='The github project',
    )
    prepare_parser.add_argument(
        '--conventional-commits',
        '-CC',
        help=(
            'Wether to use conventional commits and create '
            'the changelog directly from the git log'
        ),
        action='store_true',
    )
    prepare_parser.add_argument(
        '--conventional-commits-config',
        dest='cc_config',
        default=Path('changelog.toml'),
        type=FileType('r'),
        help="Conventional commits config file (toml), including conventions.",
    )

    release_parser = subparsers.add_parser('release')
    release_parser.set_defaults(func=release)
    release_parser.add_argument(
        '--release-version',
        help=(
            'Will release changelog as version. Must be PEP 440 compliant. '
            'default: lookup version in project definition.'
        ),
    )

    release_parser.add_argument(
        '--next-version',
        help=(
            'Sets the next PEP 440 compliant version in project definition '
            'after the release. default: set to next dev version'
        ),
    )

    release_parser.add_argument(
        '--git-remote-name',
        help='The git remote name to push the commits and tag to',
    )
    release_parser.add_argument(
        '--git-tag-prefix',
        default='v',
        help='Prefix for git tag versions. Default: %(default)s',
    )
    release_parser.add_argument(
        '--git-signing-key',
        help='The key to sign the commits and tag for a release',
    )
    release_parser.add_argument(
        '--project',
        help='The github project',
    )
    release_parser.add_argument(
        '--space',
        default='greenbone',
        help='User/Team name in github',
    )
    release_parser.add_argument(
        '--changelog',
        help=(
            'The CHANGELOG file path, defaults '
            'to CHANGELOG.md in the repository root directory'
        ),
    )
    release_parser.add_argument(
        '--conventional-commits',
        '-CC',
        help=(
            'Wether to use conventional commits and create '
            'the changelog directly from the git log'
        ),
        action='store_true',
    )

    sign_parser = subparsers.add_parser('sign')
    sign_parser.set_defaults(func=sign)
    sign_parser.add_argument(
        '--signing-key',
        default='0ED1E580',
        help='The key to sign zip, tarballs of a release. Default %(default)s.',
    )
    sign_parser.add_argument(
        '--release-version',
        help='Will release changelog as version. Must be PEP 440 compliant.',
    )
    sign_parser.add_argument(
        '--git-tag-prefix',
        default='v',
        help='Prefix for git tag versions. Default: %(default)s',
    )
    sign_parser.add_argument(
        '--project',
        help='The github project',
    )
    sign_parser.add_argument(
        '--space',
        default='greenbone',
        help='user/team name in github',
    )

    sign_parser.add_argument(
        '--passphrase',
        help=(
            'Use gpg in a headless mode e.g. for '
            'the CI and use this passphrase for signing.'
        ),
    )
    return parser


def parse(args=None) -> Tuple[str, str, Namespace]:
    parser = initialize_default_parser()
    commandline_arguments = parser.parse_args(args)
    token = os.environ['GITHUB_TOKEN'] if not args else 'TOKEN'
    user = os.environ['GITHUB_USER'] if not args else 'USER'
    return (user, token, commandline_arguments)


def main(
    shell_cmd_runner=lambda x: subprocess.run(
        x,
        shell=True,
        check=True,
        errors="utf-8",  # use utf-8 encoding for error output
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ),
    _path: Path = Path,
    _requests: requests = requests,
    _version: version = version,
    _changelog: changelog = changelog,
    leave: bool = True,
    args=None,
):
    username, token, parsed_args = parse(args)
    term = Terminal()
    _set_terminal(term)

    term.bold_info(f'pontos-release => {parsed_args.func.__name__}')

    with term.indent():
        try:
            if not parsed_args.func(
                shell_cmd_runner,
                parsed_args,
                path=_path,
                username=username,
                token=token,
                changelog_module=_changelog,
                requests_module=_requests,
                version_module=_version,
            ):
                return sys.exit(1) if leave else False
        except subprocess.CalledProcessError as e:
            if not '--passphrase' in e.cmd:
                error(f'Could not run command "{e.cmd}".')
            else:
                error('Headless signing failed.')
            out(f'Error was: {e.stderr}')
            sys.exit(1)

    return sys.exit(0) if leave else True


if __name__ == '__main__':
    main()



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import contextlib
import io

from pathlib import Path
from pontos.version.version import VersionCommand
from pontos.version.helper import VersionError

# pylint: disable=W0212


class VersionCommandTestCase(unittest.TestCase):
    def test_missing_cmd(self):
        with io.StringIO() as buf, contextlib.redirect_stdout(buf):
            VersionCommand().run()
            self.assertEqual(
                buf.getvalue(),
                'usage: version [-h] [--quiet] {verify,show,update} ...\n',
            )

    def test_missing_file(self):
        with self.assertRaises(
            VersionError, msg="Could not find whatever file."
        ):
            VersionCommand(project_file_path=Path('whatever')).run(
                args=['show']
            )



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Argument parser for pontos-github """

from argparse import ArgumentParser, Namespace
import os
from pathlib import Path
from typing import List

from pontos.github.cmds import pull_request

body_template = Path(__file__).parent / "pr_template.md"


def from_env(name: str) -> str:
    return os.environ.get(name, name)


def parse_args(
    args: List[str] = None,
) -> Namespace:
    """Parsing args for nasl-lint

    Arguments:
    args        The programm arguments passed by exec
    term        The terminal to print"""

    parser = ArgumentParser(
        description="Greenbone GitHub API.",
    )

    subparsers = parser.add_subparsers(
        title='subcommands',
        description='valid subcommands',
        required=True,
        help='additional help',
        dest='command',
    )

    # create a PR from command line
    pr_parser = subparsers.add_parser(
        'pull-request', aliases=['pr', 'PR', 'pullrequest']
    )

    pr_parser.set_defaults(func=pull_request)

    pr_parser.add_argument(
        "repo", help=("GitHub repository (owner/name) to use")
    )

    pr_parser.add_argument(
        "head",
        help=("Branch to create a pull request from"),
    )

    pr_parser.add_argument(
        "target",
        default="main",
        help="Branch as as target for the pull. Default: %(default)s",
    )

    pr_parser.add_argument(
        "title",
        help=("Title for the pull request"),
    )

    pr_parser.add_argument(
        "-b",
        "--body",
        default=body_template.read_text(encoding="utf-8"),
        help=(
            "Description for the pull request. Can be formatted in Markdown."
        ),
    )

    pr_parser.add_argument(
        "-t",
        "--token",
        default="GITHUB_TOKEN",
        type=from_env,
        help=(
            "GitHub Token to access the repository. "
            "Default looks for environment variable 'GITHUB_TOKEN'"
        ),
    )

    return parser.parse_args(args)



# -*- coding: utf-8 -*-
# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from argparse import Namespace
from pathlib import Path
import unittest
from dataclasses import dataclass
from datetime import datetime

from pontos.terminal import _set_terminal
from pontos.terminal.terminal import Terminal

from pontos import changelog


@dataclass
class StdOutput:
    stdout: bytes


class ConventionalCommitsTestCase(unittest.TestCase):
    def test_changelog_builder(self):

        _set_terminal(Terminal())

        today = datetime.today().strftime('%Y-%m-%d')

        own_path = Path(__file__).absolute().parent
        print(own_path)
        config_toml = own_path / 'changelog.toml'
        release_version = '0.0.2'
        output = f'v{release_version}.md'
        git_log = (
            '1234567 Add: foo bar\n'
            '8abcdef Add: bar baz\n'
            '8abcd3f Add bar baz\n'
            '8abcd3d Adding bar baz\n'
            '1337abc Change: bar to baz\n'
            '42a42a4 Remove: foo bar again\n'
            'fedcba8 Test: bar baz testing\n'
            'dead901 Refactor: bar baz ref\n'
            'fedcba8 Fix: bar baz fixing\n'
            'd0c4d0c Doc: bar baz documenting\n'
        )
        called = []

        def runner(cmd):
            called.append(cmd)
            return StdOutput(git_log)

        cargs = Namespace(
            current_version='0.0.1',
            next_version=release_version,
            output=output,
            space='foo',
            project='bar',
            config=config_toml,
        )
        changelog_builder = changelog.ChangelogBuilder(
            shell_cmd_runner=runner,
            args=cargs,
        )
        expected_output = f"""# Changelog

All notable changes to this project will be documented in this file.

## [0.0.2] - {today}

## Added
* foo bar [1234567](https://github.com/foo/bar/commit/1234567)
* bar baz [8abcdef](https://github.com/foo/bar/commit/8abcdef)

## Removed
* foo bar again [42a42a4](https://github.com/foo/bar/commit/42a42a4)

## Changed
* bar to baz [1337abc](https://github.com/foo/bar/commit/1337abc)

## Bug Fixes
* bar baz fixing [fedcba8](https://github.com/foo/bar/commit/fedcba8)

## Documentation
* bar baz documenting [d0c4d0c](https://github.com/foo/bar/commit/d0c4d0c)

## Refactor
* bar baz ref [dead901](https://github.com/foo/bar/commit/dead901)

## Testing
* bar baz testing [fedcba8](https://github.com/foo/bar/commit/fedcba8)

[0.0.2]: https://github.com/foo/bar/compare/0.0.1...0.0.2"""
        output_file = changelog_builder.create_changelog_file()
        print(output_file.read_text(encoding='utf-8'))
        self.assertEqual(
            output_file.read_text(encoding='utf-8'), expected_output
        )

        self.assertIn(
            'git log "$(git describe --tags --abbrev=0)..HEAD" --oneline',
            called,
        )

        output_file.unlink()

    def test_changelog_builder_no_commits(self):

        _set_terminal(Terminal())

        own_path = Path(__file__).absolute().parent
        print(own_path)
        config_toml = own_path / 'changelog.toml'
        release_version = '0.0.2'
        output = f'v{release_version}.md'
        git_log = '1234567 foo bar\n8abcdef bar baz\n'
        called = []

        def runner(cmd):
            called.append(cmd)
            return StdOutput(git_log)

        cargs = Namespace(
            current_version='0.0.1',
            next_version=release_version,
            output=output,
            space='foo',
            project='bar',
            config=config_toml,
        )
        changelog_builder = changelog.ChangelogBuilder(
            shell_cmd_runner=runner,
            args=cargs,
        )

        with self.assertRaises(SystemExit):
            output_file = changelog_builder.create_changelog_file()

            self.assertIsNone(output_file)
            self.assertIn(
                'git log "$(git describe --tags --abbrev=0)..HEAD" --oneline',
                called,
            )

    def test_changelog_builder_no_conventional_commits(self):

        _set_terminal(Terminal())

        own_path = Path(__file__).absolute().parent
        config_toml = own_path / 'changelog.toml'
        release_version = '0.0.2'
        output = f'v{release_version}.md'
        git_log = None
        called = []

        def runner(cmd):
            called.append(cmd)
            return StdOutput(git_log)

        cargs = Namespace(
            current_version='0.0.1',
            next_version=release_version,
            output=output,
            space='foo',
            project='bar',
            config=config_toml,
        )
        changelog_builder = changelog.ChangelogBuilder(
            shell_cmd_runner=runner,
            args=cargs,
        )

        with self.assertRaises(SystemExit):
            output_file = changelog_builder.create_changelog_file()

            self.assertIsNone(output_file)
            self.assertIn(
                'git log "$(git describe --tags --abbrev=0)..HEAD" --oneline',
                called,
            )



# -*- coding: utf-8 -*-
# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# pylint: disable=C0413,W0108

import os
import unittest

from dataclasses import dataclass
from pathlib import Path
from unittest.mock import MagicMock, patch

import requests

from pontos.release.helper import version
from pontos import release, changelog


@dataclass
class StdOutput:
    stdout: bytes


class PrepareTestCase(unittest.TestCase):
    def setUp(self) -> None:
        os.environ['GITHUB_TOKEN'] = 'foo'
        os.environ['GITHUB_USER'] = 'bar'

    def test_prepare_successfully(self):
        fake_path_class = MagicMock(spec=Path)
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')
        args = [
            'prepare',
            '--release-version',
            '0.0.1',
        ]
        runner = lambda x: StdOutput('')
        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertTrue(released)

    def test_prepare_calendar_successfully(self):
        fake_path_class = MagicMock(spec=Path)
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')

        args = [
            'prepare',
            '--calendar',
        ]
        runner = lambda x: StdOutput('')
        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertTrue(released)

    def test_use_git_signing_key_on_prepare(self):
        fake_path_class = MagicMock(spec=Path)
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')

        args = [
            'prepare',
            '--git-signing-key',
            '0815',
            '--release-version',
            '0.0.1',
        ]
        called = []

        def runner(cmd):
            called.append(cmd)
            return StdOutput('')

        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertTrue(released)
        self.assertIn(
            "git commit -S0815 --no-verify -m 'Automatic release to 0.0.1'",
            called,
        )
        self.assertIn(
            "git tag -u 0815 v0.0.1 -m 'Automatic release to 0.0.1'", called
        )

    def test_fail_if_tag_is_already_taken(self):
        fake_path_class = MagicMock(spec=Path)
        fake_version = MagicMock(spec=version)
        fake_changelog = MagicMock(spec=changelog)

        args = [
            'prepare',
            '--release-version',
            '0.0.1',
            '--project',
            'bla',
        ]

        called = []

        def runner(cmd):
            called.append(cmd)
            return StdOutput('v0.0.1'.encode())

        with self.assertRaises(SystemExit):
            release.main(
                shell_cmd_runner=runner,
                _path=fake_path_class,
                _version=fake_version,
                _changelog=fake_changelog,
                leave=False,
                args=args,
            )

            self.assertIn('git tag v0.0.1 is already taken', called)

    def test_not_release_when_no_project_found(self):
        fake_path_class = MagicMock(spec=Path)
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (False, '')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')

        args = [
            'prepare',
            '--release-version',
            '0.0.1',
        ]
        runner = lambda x: StdOutput('')
        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertFalse(released)

    def test_not_release_when_updating_version_fails(self):
        fake_path_class = MagicMock(spec=Path)
        fake_requests = MagicMock(spec=requests)
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (False, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')

        args = [
            'prepare',
            '--release-version',
            '0.0.1',
        ]
        runner = lambda x: StdOutput('')
        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _requests=fake_requests,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertFalse(released)

    @patch('pontos.changelog.changelog')
    def test_prepare_coventional_commits(self, changelog_mock):
        # fake_path_class = MagicMock(spec=Path)
        fake_requests = MagicMock(spec=requests)
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')

        own_path = Path(__file__).absolute().parent
        release_file = own_path.parent.parent / '.release.md'

        builder = changelog_mock.ChangelogBuilder.return_value
        builder.create_changelog_file.return_value = own_path / 'v1.2.3.md'

        args = [
            'prepare',
            '--release-version',
            '1.2.3',
            '-CC',
        ]
        runner = lambda x: StdOutput('')
        released = release.main(
            shell_cmd_runner=runner,
            _path=Path,
            _requests=fake_requests,
            _version=fake_version,
            _changelog=changelog_mock,
            leave=False,
            args=args,
        )
        self.assertTrue(released)

        expected_release_content = """## [21.8.1] - 2021-08-23

## Added

* Need for commits. [1234567](https://github.com/foo/bar/commit/1234567)

## Changed

* fooooo. [1234568](https://github.com/foo/bar/commit/1234568)

[21.8.1]: https://github.com/y0urself/test_workflows/compare/21.8.0...21.8.1"""

        self.assertEqual(
            release_file.read_text(encoding='utf-8'), expected_release_content
        )

        release_file.unlink()



# -*- coding: utf-8 -*-
# pontos/release/release.py
# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from argparse import Namespace
import json

from pathlib import Path
from typing import Callable

import requests

from pontos.terminal import error, info, out

from .helper import (
    download,
    get_current_version,
    get_project_name,
    upload_assets,
    download_assets,
)


def sign(
    shell_cmd_runner: Callable,
    args: Namespace,
    *,
    path: Path,
    username: str,
    token: str,
    requests_module: requests,
    **_kwargs,
) -> bool:

    project: str = (
        args.project
        if args.project is not None
        else get_project_name(shell_cmd_runner)
    )
    space: str = args.space
    git_tag_prefix: str = args.git_tag_prefix
    release_version: str = (
        args.release_version
        if args.release_version is not None
        else get_current_version()
    )
    signing_key: str = args.signing_key

    headers = {'Accept': 'application/vnd.github.v3+json'}

    git_version: str = f'{git_tag_prefix}{release_version}'

    base_url = (
        f"https://api.github.com/repos/{space}/{project}"
        f"/releases/tags/{git_version}"
    )

    response = requests_module.get(
        base_url,
        headers=headers,
    )
    if response.status_code != 200:
        error(
            f"Wrong response status code {response.status_code} for request "
            f"{base_url}"
        )
        out(json.dumps(response.text, indent=4, sort_keys=True))
        return False

    zipball_url = (
        f"https://github.com/{space}/{project}/archive/refs/"
        f"tags/{git_version}.zip"
    )
    github_json = json.loads(response.text)
    zip_path = download(
        zipball_url,
        f"{project}-{release_version}.zip",
        path=path,
        requests_module=requests_module,
    )
    tarball_url = (
        f"https://github.com/{space}/{project}/archive/refs/"
        f"tags/{git_version}.tar.gz"
    )
    tar_path = download(
        tarball_url,
        f"{project}-{release_version}.tar.gz",
        path=path,
        requests_module=requests_module,
    )

    file_paths = [zip_path, tar_path]

    asset_paths = download_assets(
        github_json.get('assets_url'),
        path=path,
        requests_module=requests_module,
    )

    file_paths.extend(asset_paths)

    for file_path in file_paths:
        info(f"Signing {file_path}")

        if args.passphrase:
            shell_cmd_runner(
                f"gpg --pinentry-mode loopback --default-key {signing_key}"
                f" --yes --detach-sign --passphrase {args.passphrase}"
                f" --armor {file_path}"
            )
        else:
            shell_cmd_runner(
                f"gpg --default-key {signing_key} --yes --detach-sign --armor "
                f"{file_path}"
            )

    return upload_assets(
        username,
        token,
        file_paths,
        github_json,
        path=path,
        requests_module=requests_module,
    )
